-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-- Written by Kimberly L. Tripp, SQLskills.com
-- All Rights Reserved.
-- 
-- For more scripts, sample code and additional training resources,
-- go to http://www.SQLskills.com
--	Immerse yourself in SQL Server
--
-- Disclaimer - Thoroughly test this script, execute at your own risk.
-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

-- For simplicity and SO that you can reproduce this example later, we
-- will use the pubs database!

-- First, you'll need to change it's recovery model
ALTER DATABASE pubs
	SET RECOVERY FULL
go

-- Next, you'll need to begin your "recovery strategy setup" by performing
-- some form of starting backup. In this case, we'll choose a full backup. 
-- NOTE: File or filegroup backups are also possible as a starting point
-- for recovery.
BACKUP DATABASE pubs
	TO DISK = N'C:\pubsbackup.bak'
	WITH INIT
go

-- Now we can begin to perform transactions as well as manage those transactions
-- through transaction log backups.

USE pubs
go

-- Again, for simplicity - we'll make changes to the titles table in order to 
-- setup the same set of transactions as in the slide: "Transaction Recovery & Checkpoints"

SELECT t.title_id 	AS [Title ID]
	, t.title	AS [Title]
FROM dbo.titles AS t
ORDER BY [Title ID]
go

-- This is transaction 1 from slide titled: "Transaction Recovery & Checkpoints"
UPDATE titles
	SET title = 'T1 commit b4 checkpoint - committed'
	WHERE title_id = 'BU1032'
go

-- This is transaction 3 from slide titled: "Transaction Recovery & Checkpoints"
BEGIN TRANSACTION
UPDATE titles
	SET title = 'T2 change b4 checkpoint - NOT committed'
	WHERE title_id = 'BU1111'

-- go to another window and run a checkpoint in this database

-- shutdown the server with nowait or power it off... (power failure, etc.)

-- for the first test you should delete the data file

-- restart server

-- check databases... pubs should be suspect
sp_helpdb
go
-- pubs won't even be listed?

-- can we get in?
USE pubs
go

-- No! You should get ERROR 945

-- NOW - backup the log with NO_TRUNCATE.
BACKUP LOG pubs
	TO DISK = N'C:\pubsbackup.bak'
	WITH NOINIT, NO_TRUNCATE

-- The NO_TRUNCATE option can be used to backup the transaction log for a database
-- which is suspect due to corruption in the data portion of the database.

-- Once the tail of the transaction log is backed up then you can recover
-- the database by restoring the backup and applying the transaction log backup.

-- switch to another database and make sure that no query windows are connected to pubs

USE master
go

RESTORE DATABASE pubs
	FROM DISK = N'C:\pubsbackup.bak'
	WITH FILE = 1
go

RESTORE LOG pubs
	FROM DISK = N'C:\pubsbackup.bak'
	WITH FILE = 1
go

USE pubs
go

SELECT t.title_id 	AS [Title ID]
	, t.title	AS [Title]
FROM dbo.titles AS t
ORDER BY [Title ID]

-- Notice that BU1111 is back to the correct title! This "partial" and "inflight" 
-- transaction was rolled back through correct application of the transaction
-- log backup

-- Next, we'll recreate the scenario and trash the log!

BACKUP DATABASE pubs
	TO DISK = N'C:\pubsbackup.bak'
	WITH INIT
-- this will create a new backup of the database and overwrite the backups that are there.
go

-- Let's re-create transaction 3 from slide titled: "Transaction Recovery & Checkpoints"
BEGIN TRANSACTION
UPDATE titles
	SET title = 'T2 change b4 checkpoint - NOT committed'
	WHERE title_id = 'BU1111'

-- go to another window and run a checkpoint in this database

-- shutdown the server with nowait or power it off... (power failure, etc.)

-- for this test we will delete the LOG file

-- Once the server is restarted the database will once again be marked suspect...

-- Can you backup the log?
BACKUP LOG pubs
	TO DISK = N'C:\pubsbackup.bak'
	WITH NOINIT, NO_TRUNCATE

-- NO!

-- What if we have nothing? What if we REALLY want to get into that database - can we?
-- disclaimer - these next few steps are NOT generally recommended.

USE master
go

sp_configure 'allow updates', 1
go
reconfigure with override
go

update master.dbo.sysdatabases
	set status = 32768
	where name = 'pubs'
go

sp_configure 'allow updates', 0
go
reconfigure with override
go

-- Now can we get in?
USE pubs
go

-- Sure, but what's the state of the database?

SELECT t.title_id 	AS [Title ID]
	, t.title	AS [Title]
FROM dbo.titles AS t
ORDER BY [Title ID]
go

-- That 'partial' and 'inflight' transaction is "in" the database... 
-- You CANNOT guarantee that all of your data is consistent?
-- SQL Server can't - at least not without the log. This is why Microsoft
-- says that when a database is suspect you should begin your recovery process...

-- Your best step now would be to transfer and verify all of your data to a NEW
-- database. Once all of the data is there you should recreate all of your indexes.

-- Would have been A LOT easier if you had access to the log and/or if your last log 
-- backup were recent enough that you could have just used that. Using transaction
-- log backups ensure data consistency!!!

-- HAVE A BACKUP STRATEGY! 
--	TEST YOUR BACKUP STRATEGY!


















-- What's your recovery?
-- First see if you can access the log?
-- Backup the tail of the log (when possible)
-- Recover from backups...

-- So - if SQL Server had the log then we wouldn't see this partial transaction...
-- let's see what would have happened if we hadn't had media failure, just power
-- failure?

-- shutdown the server and delete the data but put back the log...
sp_configure 'allow updates', 1
go
reconfigure with override
go

update master.dbo.sysdatabases
	set status = 16
	where name = 'pubs'
go

sp_configure 'allow updates', 0
go
reconfigure with override
go

shutdown with nowait
go

use pubs
go

-- same problem - error 945

-- pubs is suspect but we have the log from which we can recover!
-- How?
-- backup the tail of the transaction log

-- First, can we get any data?
SELECT t.title_id 	AS [Title ID]
	, t.title		AS [Title]
FROM dbo.titles AS t
ORDER BY [Title ID]
go

-- Can we do anything?
-- Backup the tail of the transaction log!

BACKUP LOG pubs
	TO DISK = N'C:\pubsbackup.bak'
	WITH NOINIT, NO_TRUNCATE

-- Sure, but what's the state of the database?

SELECT t.title_id 	AS [Title ID]
	, t.title		AS [Title]
FROM dbo.titles AS t
ORDER BY [Title ID]
go

-- Restore the original version of the database back to the start:
RESTORE DATABASE pubs
	FROM DISK = N'C:\pubsbackup.bak'
	WITH RECOVERY, REPLACE 